#!/usr/bin/env python3
#
# This file ist documented in the ./README.md file. Please read it for deteiled explanations.
# Usage: `python3 camera_calibration.py` or mark as executable and call directly

# * ----- Package imports: ----- *
import glob
import pathlib

import cv2 as cv
import numpy as np
import yaml


# * ----- Configurations: ----- *
# * change the following configurations to fit your usecase:

# define the grid pattern to look fore
PATTERN = (7,7)



# * ----- Script code: ----- *
if __name__ == "__main__":

    # termination criteria
    criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 30, 0.001)

    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
    objp = np.zeros((PATTERN[0]*PATTERN[1],3), np.float32)
    objp[:,:2] = np.mgrid[0:PATTERN[0],0:PATTERN[1]].T.reshape(-1,2)

    # Arrays to store object points and image points from all the images.
    objpoints = [] # 3d point in real world space
    imgpoints = [] # 2d points in image plane.

    # get all images in current directory
    folder = pathlib.Path(__file__).parent.resolve()
    images = glob.glob(f'{folder}/*.png')

    # loop over all images:
    for fname in images:
        print(fname)
        img = cv.imread(fname)
        gray = cv.cvtColor(img, cv.COLOR_BGR2GRAY)

        # Find the chess board corners
        ret, corners = cv.findChessboardCorners(gray, PATTERN, None, flags=cv.CALIB_CB_ADAPTIVE_THRESH)

        # If found, add object points, image points (after refining them)
        if ret == True:
            print("valid")
            objpoints.append(objp)
            corners2 = cv.cornerSubPix(gray,corners, (11,11), (-1,-1), criteria)
            imgpoints.append(corners)

            # Draw and display the corners
            # cv.drawChessboardCorners(img, (7,6), corners2, ret)
            # cv.imshow('img', img)
            # cv.waitKey()

    # get calibration parameters:
    ret, mtx, dist, rvecs, tvecs = cv.calibrateCamera(objpoints, imgpoints, gray.shape[::-1], None, None)


    # --- test ---
    test_image = images[0]
    img = cv.imread(test_image)
    cv.namedWindow("original", cv.WINDOW_NORMAL)
    cv.imshow('original', img)

    h, w = img.shape[:2]
    newcameramtx, roi = cv.getOptimalNewCameraMatrix(mtx, dist, (w,h), 1, (w,h))

    # undistort
    dst = cv.undistort(img, mtx, dist, None, newcameramtx)
    cv.namedWindow("undistorted", cv.WINDOW_NORMAL)
    cv.imshow('undistorted', dst)

    # crop the image
    x, y, w, h = roi
    dst = dst[y:y+h, x:x+w]
    cv.namedWindow("croped", cv.WINDOW_NORMAL)
    cv.imshow('croped', dst)


    print(
        yaml.dump(
            {"intrinsic": {"matrix": mtx.tolist(), "distortion": dist.tolist()[0]}},
            default_flow_style=None,
        )
    )

    # calculate re-projection error
    mean_error = 0
    for i in range(len(objpoints)):
        imgpoints2, _ = cv.projectPoints(objpoints[i], rvecs[i], tvecs[i], mtx, dist)
        error = cv.norm(imgpoints[i], imgpoints2, cv.NORM_L2)/len(imgpoints2)
        mean_error += error
    print(f"total error: {mean_error/len(objpoints)}")

    cv.waitKey()
