#include <ros/ros.h>
#include <sensor_msgs/Image.h>
#include <opencv2/opencv.hpp>
#include <cv_bridge/cv_bridge.h>
#include <list>

#include "lane_markings_detection.hpp"


// global variables needed by callback
ros::Publisher* image_pub;


/**
 * @brief runns a canny-edge detector on the image and returns an
 * orientation classification for each edge pixel
 *
 * @param image the image to analyse
 * @return DataImage<uint8_t>
 *      A image of itentical dimentions with edge-pixels marked
 *      with their classification value
 */
DataImage<uint8_t> edgeDetectionClassification(Image image) {
    DataImage<int> gradients(image.width, image.height);
    DataImage<float> angles(image.width, image.height);
    DataImage<uint8_t> canny_edges(image.width, image.height);

    // Compute gradient magnitude and orientation
    for (int u=1; u < image.height-1; u++) {
        for (int v=1; v < image.width-1; v++) {

            int dx=0, dy=0;
            // std::vector<int>::iterator sobel_x = SOBEL_X.begin();
            // std::vector<int>::iterator sobel_y = SOBEL_Y.begin();
            for (int y=0; y<3; y++){
                for (int x=0; x<3; x++){
                    dx += SOBEL_X[y*3+x] * image[pnt(v+y-1,u+x-1)];
                    dy += SOBEL_Y[y*3+x] * image[pnt(v+y-1,u+x-1)];
                }
            }
            gradients[pnt(v,u)] = static_cast<int>(sqrt(dx*dx + dy*dy));
            angles[pnt(v,u)] = static_cast<float>(atan2(dy,dx) / 3.1415 * 180.0 );
        }
    }

    // Non-maximum suppression
    // and classification
    for (int u=1; u < image.height-1; u++) {
        for (int v=1; v < image.width-1; v++) {
            uint8_t negative = 0;

            int grad = gradients[pnt(v,u)];
            int arc = angles[pnt(v,u)];

            if (arc < 0)
                negative = 0x10;
            arc = fabsf(arc);

            if (arc <= 22.5f || arc > 157.5f ) {
                canny_edges[pnt(v,u)] = negative | V;
                if (grad < gradients[pnt(v,u-1)] || grad < gradients[pnt(v,u+1)])
                    gradients[pnt(v,u)] = 0;
            } else if ( ( !negative && arc < 67.5f ) || ( negative && 112.5f <= arc ) ) {
                canny_edges[pnt(v,u)] = negative | D1;
                if (grad < gradients[pnt(v-1,u-1)] || grad < gradients[pnt(v+1,u+1)])
                    gradients[pnt(v,u)] = 0;
            } else if ( ( negative && arc < 67.5f ) || ( !negative && 112.5f <= arc ) ) {
                canny_edges[pnt(v,u)] = negative | D2;
                if (grad < gradients[pnt(v-1,u+1)] || grad < gradients[pnt(v+1,u-1)])
                    gradients[pnt(v,u)] = 0;
            } else if ( 67.5f <= arc && arc < 112.5f ) {
                canny_edges[pnt(v,u)] = negative | H;
                if (grad < gradients[pnt(v-1,u)] || grad < gradients[pnt(v+1,u)])
                    gradients[pnt(v,u)] = 0;
            } else {
                ROS_DEBUG("got strange angle: %i", arc);
            }
        }
    }

    // thresholding
    uint8_t T1 = 40;
    for (int u=1; u < image.height-1; u++)
        for (int v=1; v < image.width-1; v++) {
            pnt p(v,u);
            canny_edges[p] = gradients[p] < T1 ? 0 : canny_edges[p];
        }
    return canny_edges;
}


/**
 * @brief this function gets called back on every image that is published
 *
 * @param image_msg The published image, gets passed in by ROS
 */
void callback_image(sensor_msgs::Image image_msg) {

    ros::Time starttime = ros::Time::now();
    std::list<Line> all_lines;
    std::list<Line> left_D1_edges;
    std::list<Line> right_D1_edges;
    std::list<Line> left_D2_edges;
    std::list<Line> right_D2_edges;
    std::list<LineMarking> markings_found;
    Image image(image_msg);


    // run canny-edge detector and classify edges
    DataImage<uint8_t> classified_edges = edgeDetectionClassification(image);


    // find starting points and follow continues lines
    // creating Line objects
    for (int u=1; u < classified_edges.height-1; u++) {
        for (int v=1; v < classified_edges.width-1; v++) {
            if (!classified_edges[pnt(v,u)])
                continue;

            // get only classification without direction:
            uint8_t clsif_org = classified_edges[pnt(v,u)];
            uint8_t clsif = 0x0f & clsif_org;
            bool has_neighbour = false;

            switch (clsif) {
                case V:
                    if (
                        V == (0x0f & classified_edges[pnt(v,u-1)])   ||
                        V == (0x0f & classified_edges[pnt(v-1,u-1)]) ||
                        V == (0x0f & classified_edges[pnt(v+1,u-1)])
                    )
                        has_neighbour = true;
                    break;
                case D1:
                    if (
                        D1 == (0x0f & classified_edges[pnt(v,u-1)]) ||
                        D1 == (0x0f & classified_edges[pnt(v+1,u)]) ||
                        D1 == (0x0f & classified_edges[pnt(v+1,u-1)])
                    )
                        has_neighbour = true;
                    break;
                case D2:
                    if (
                        D2 == (0x0f & classified_edges[pnt(v-1,u)]) ||
                        D2 == (0x0f & classified_edges[pnt(v,u-1)]) ||
                        D2 == (0x0f & classified_edges[pnt(v-1,u-1)])
                    )
                        has_neighbour = true;
                    break;
                case H:
                    if (
                        H == (0x0f & classified_edges[pnt(v-1,u)])   ||
                        H == (0x0f & classified_edges[pnt(v-1,u-1)]) ||
                        H == (0x0f & classified_edges[pnt(v-1,u+1)])
                    )
                        has_neighbour = true;
                    break;
            }
            if ( has_neighbour )
                continue;

            // no neighbour found means start pixel
            pnt start(v,u);
            // follow line to its end
            do {
                classified_edges[pnt(v,u)] = 0;
                has_neighbour = false;
                switch (clsif) {
                    case V:
                        if ( V == (0x0f & classified_edges[pnt(v,u+1)]) ) {
                            u++;
                            has_neighbour = true;
                        } else if ( V == (0x0f & classified_edges[pnt(v+1,u+1)]) ) {
                            u++;
                            v++;
                            has_neighbour = true;
                        } else if ( V == (0x0f & classified_edges[pnt(v-1,u+1)]) ) {
                            u++;
                            v--;
                            has_neighbour = true;
                        }
                        break;
                    case D1:
                        if ( D1 == (0x0f & classified_edges[pnt(v,u+1)]) ) {
                            u++;
                            has_neighbour = true;
                        } else if ( D1 == (0x0f & classified_edges[pnt(v-1,u)]) ) {
                            v--;
                            has_neighbour = true;
                        } else if ( D1 == (0x0f & classified_edges[pnt(v-1,u+1)]) ) {
                            u++;
                            v--;
                            has_neighbour = true;
                        }
                        break;
                    case D2:
                        if ( D2 == (0x0f & classified_edges[pnt(v+1,u)]) ) {
                            v++;
                            has_neighbour = true;
                        } else if ( D2 == (0x0f & classified_edges[pnt(v,u+1)]) ) {
                            u++;
                            has_neighbour = true;
                        } else if ( D2 == (0x0f & classified_edges[pnt(v+1,u+1)]) ) {
                            u++;
                            v++;
                            has_neighbour = true;
                        }
                        break;
                    case H:
                        if ( H == (0x0f & classified_edges[pnt(v+1,u)]) ) {
                            v++;
                            has_neighbour = true;
                        } else if ( H == (0x0f & classified_edges[pnt(v+1,u+1)]) ) {
                            u++;
                            v++;
                            has_neighbour = true;
                        } else if ( H == (0x0f & classified_edges[pnt(v+1,u-1)]) ) {
                            u--;
                            v++;
                            has_neighbour = true;
                        }
                        break;
                }
            } while ( has_neighbour );

            Line l(start, pnt(v,u), clsif_org);
            if ( l.length > 5.0 ) {
                all_lines.push_back(l);
                if (clsif == D1) {
                    if ( clsif_org & 0x10 ) {
                        right_D1_edges.push_back(l);
                    } else {
                        left_D1_edges.push_back(l);
                    }
                }
                if (clsif == D2) {
                    if ( clsif_org & 0x10 ) {
                        right_D2_edges.push_back(l);
                    } else {
                        left_D2_edges.push_back(l);
                    }
                }
            }

            // reset to original coords, so for loop continues where it left off
            u = start.y;
            v = start.x;
        }
    }


    for(const Line& a : left_D1_edges) {
        for(auto it = right_D1_edges.begin(); it != right_D1_edges.end(); it++ ) {
            const Line b = *it;
            if (
                ( a.start.y - 10 < b.start.y && b.start.y < a.start.y + 10 ) &&
                ( a.start.x < b.start.x && b.start.x < a.start.x + 35 )
            ) {
                markings_found.push_back(LineMarking(a,b, 0x80));
                left_D1_edges.erase(it);
                break;
            }
        }
    }

    for(const Line& a : left_D2_edges) {
        for(auto it = right_D2_edges.begin(); it != right_D2_edges.end(); it++ ) {
            const Line b = *it;
            if (
                ( a.start.y - 20 < b.start.y && b.start.y < a.start.y + 20 ) &&
                ( a.start.x < b.start.x && b.start.x < a.start.x + 30 )
            ) {
                markings_found.push_back(LineMarking(a,b, 0x80));
                left_D1_edges.erase(it);
                break;
            }
        }
    }


    cv::Mat displayLines = cv::Mat::zeros(cv::Size(image.width, image.height), CV_8U);


    for(const Line& l : all_lines) {
        cv::line(displayLines, l.start, l.end, cv::Scalar(40));
    }

    for(const LineMarking& a : markings_found) {

        std::vector<std::vector<cv::Point>> contourVec;
        contourVec.push_back(a.contour);
        cv::drawContours(displayLines, contourVec, 0, cv::Scalar(128));

        cv::line(displayLines, a.middle->start, a.middle->end, cv::Scalar(255));
    }

    cv_bridge::CvImage image_displayLines(std_msgs::Header(), "mono8", displayLines);
    image_pub->publish(image_displayLines.toImageMsg());

    ros::Time endtime = ros::Time::now();
    ros::Duration time_spend = endtime - starttime;
    ROS_DEBUG("run took: %fs", time_spend.toSec());
}



// node main loop
int main(int argc, char **argv) {
    ros::init(argc, argv, "laneMarkingDetection");

    ros::NodeHandle nh;
    ros::NodeHandle private_nh("~");

    /*
     * subscribe to camera image creation event
     */
    ros::Subscriber sub = nh.subscribe("/img/gray", 10, callback_image);

    /*
     * advertise publisher topics
     */
    ros::Publisher image_publisher = private_nh.advertise<sensor_msgs::Image>("/img/temp", 2);
    image_pub = &image_publisher;

    /*
     * start the node
     */
    ros::console::set_logger_level(ROSCONSOLE_DEFAULT_NAME, ros::console::levels::Debug);
    ros::spin();

    return 0;
}
