import cv2
import numpy as np
from math import atan2, inf, nan, sqrt, pi
from copy import copy
from time import perf_counter, strftime


sobel_x = np.array([[-1, 0, +1], [-2, 0, +2], [-1, 0, +1]])
sobel_y = np.array([[-1, -2, -1], [0, 0, 0], [+1, +2, +1]])

V = 0x01
D1 = 0x02
D2 = 0x04
H = 0x08


class Line:
    start: tuple[int, int]
    end: tuple[int, int]
    length: float
    angle: float
    info: np.uint8
    color = [255,255,255]

    def __init__(
        self, start: tuple[int, int], end: tuple[int, int], info: np.uint8 = 255
    ) -> None:
        if start[0] < end[0]:
            self.start = start
            self.end = end
        else:
            self.start = end
            self.end = start

        dx = end[0] - start[0]
        dy = end[1] - start[1]
        self.length = sqrt(dx * dx + dy * dy)
        self.angle = atan2(dy, dx) / pi * 180

        if info <= 0x1f:
            self.info = info
            self.color[0] = 255 if info & 0x10 else 0
            self.color[1] = 255 if info & 0x07 else 0
            self.color[2] = 255 if info & 0x0b else 0
            self.color = tuple(self.color)




class LineMarking:
    contour: list[tuple]
    middle: Line

    def __init__(self, a: Line, b: Line, orientation: str) -> None:
        self.contour = np.array([
            a.start, a.end, b.end, b.start
        ])
        self.middle = Line(
            (
                int((a.start[0] + b.start[0])/2),
                int((a.start[1] + b.start[1])/2),
            ),
            (
                int((a.end[0] + b.end[0])/2),
                int((a.end[1] + b.end[1])/2),
            ),
        )

lines: list[Line] = []
markings_found: list[LineMarking] = []


img = cv2.imread("./image.png")
img = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)


starttime = perf_counter()

img = cv2.GaussianBlur(img, (3,3), 1.5)
canny = cv2.Canny(img, 180, 40)


pixel_info = np.zeros(img.shape[:2], dtype=np.uint8)

# get classification for evey edge pixel
for (u, v), e in np.ndenumerate(canny[1:-1, 1:-1]):
    if not e:
        continue

    u += 1
    v += 1

    nh = img[u - 1 : u + 2, v - 1 : v + 2]
    dx = np.sum(nh * sobel_x)
    dy = np.sum(nh * sobel_y)
    arc = atan2(dy, dx) / pi * 180

    if arc < 0:
        pixel_info[u, v] |= 0x10
    arc = abs(arc)

    if arc >= 157.5 or 22.5 > arc:
        pixel_info[u, v] |= V
    elif 22.5 <= arc < 67.5:
        pixel_info[u, v] |= D1 if not pixel_info[u, v] else D2
    elif 67.5 <= arc < 112.5:
        pixel_info[u, v] |= H
    elif 112.5 <= arc < 157.5:
        pixel_info[u, v] |= D2 if not pixel_info[u, v] else D1

# find starting points and follow cintinues lines
# createing Line objects
for (u, v), e in np.ndenumerate(pixel_info[1:-1, 1:-1]):
    if not e:
        continue

    u += 1
    v += 1
    info = copy(e)
    e &= 0x0f

    if e == V:
        relevant_nh = np.array([[-1, 0], [-1, -1], [-1, 1]])
    elif e == D1:
        relevant_nh = np.array([[0, 1], [-1, 0], [-1, 1]])
    elif e == H:
        relevant_nh = np.array([[0, -1], [-1, -1], [1, -1]])
    elif e == D2:
        relevant_nh = np.array([[-1, 0], [0, -1], [-1, -1]])

    for x, y in relevant_nh:
        if e == pixel_info[u + x, v + y] & 0x0f:
            break

    else:  # no neighbour found means start pixel
        start = (u, v)
        relevant_nh *= -1

        while True:
            pixel_info[u, v] = 0
            for x, y in relevant_nh:
                if e == pixel_info[u + x, v + y] & 0x0f:
                    u, v = u + x, v + y
                    break
            else:  # no more neighbours
                break

        l = Line(start, (u, v), info)
        if l.length > 5:
            lines.append(l)



# group edges by class and find pairs
# creating LineMarking objects
left_D1_edges = [e for e in lines if e.info & 0x0f == D1 and not e.info & 0x10]
right_D1_edges = [e for e in lines if e.info & 0x0f == D1 and e.info & 0x10]
for a in left_D1_edges:
    for b in right_D1_edges:
        if (
            a.start[0] - 20 < b.start[0] < a.start[0] + 20 and
            a.start[1] < b.start[1] < a.start[1] + 30 #and
            #a.end[0] - 20 < b.end[0] < a.end[0] + 20 and
            #a.end[1] < b.end[1] < a.end[1] + 20
        ):
            markings_found.append(LineMarking(a,b, "left"))
            right_D1_edges.remove(b)
            break

left_D2_edges = [e for e in lines if e.info & 0x0f == D2 and not e.info & 0x10]
right_D2_edges = [e for e in lines if e.info & 0x0f == D2 and e.info & 0x10]
for a in left_D2_edges:
    for b in right_D2_edges:
        if (
            a.start[1] - 30 < b.start[1] < a.start[1] + 30 and
            a.start[0] < b.start[0] < a.start[0] + 30 #and
            #a.end[0] - 20 < b.end[0] < a.end[0] + 20 and
            #a.end[1] < b.end[1] < a.end[1] + 20
        ):
            markings_found.append(LineMarking(a,b, "right"))
            right_D2_edges.remove(b)
            break


endtime = perf_counter()

print(f"runtimetime: {endtime-starttime:.4f}s")


# display detected markings:
out = np.zeros(shape=img.shape + tuple([3]))
for mark in markings_found:
    cv2.drawContours(out, [np.fliplr(mark.contour)], 0, (255,255,0) )

    cv2.line(out, (mark.middle.start[1], mark.middle.start[0]), (mark.middle.end[1], mark.middle.end[0]), (0,255,0))

cv2.imshow("found lines", out)
cv2.waitKey()
