#include <ros/ros.h>
#include <sensor_msgs/Image.h>
#include <opencv2/opencv.hpp>
#include <cv_bridge/cv_bridge.h>
#include <list>

#include "lane_markings_detection.hpp"


// global variables needed by callback
ros::Publisher* image_pub;


/**
 * @brief runns a canny-edge detector on the image and returns an
 * orientation classification for each edge pixel
 *
 * @param image the image to analyse
 * @return DataImage<uint8_t>
 *      A image of itentical dimentions with edge-pixels marked
 *      with their classification value
 */
cv::Mat edgeDetectionClassification(cv::Mat image) {

    cv::Mat canny;
    cv::Canny(image, canny, 180, 50);

    cv::Mat classified_edges = cv::Mat::zeros(cv::Size(image.cols,image.rows), CV_8U);
    for (int u=1; u < image.rows-1; u++) {
        for (int v=1; v < image.cols-1; v++) {
            if ( ! canny.at<uint8_t>(u,v) )
                continue;


            uint8_t e;
            int dx=0, dy=0;
            for (int y=0; y<3; y++) {
                for (int x=0; x<3; x++) {
                    e = image.at<uint8_t>(u+y-1,v+x-1);
                    dx += SOBEL_X[y*3+x] * e;
                    dy += SOBEL_Y[y*3+x] * e;
                }
            }
            double arc = atan2(dy,dx) / 3.1415 * 180.0;

            uint8_t clsif = 0;
            if (arc < 0)
                clsif = 0x10;
            arc = fabsf(arc);

            if (arc <= 22.5f || arc > 157.5f ) {
                clsif |= V;
            } else if ( 67.5f <= arc && arc < 112.5f ) {
                clsif |= H;
            } else if ( ( !clsif && arc < 67.5f ) || ( clsif && 112.5f <= arc ) ) {
                clsif |= D1;
            } else if ( ( clsif && arc < 67.5f ) || ( !clsif && 112.5f <= arc ) ) {
                clsif |= D2;
            } else {
                ROS_DEBUG("got strange angle: %.2f", arc);
                continue;
            }
            classified_edges.at<uint8_t>(u,v) = clsif;
        }
    }
    return classified_edges;
}



void callback_image(sensor_msgs::Image image_msg) {

    ros::Time starttime = ros::Time::now();
    std::list<Line> all_lines;
    std::list<Line> left_D1_edges;
    std::list<Line> right_D1_edges;
    std::list<Line> left_D2_edges;
    std::list<Line> right_D2_edges;
    std::list<LineMarking> markings_found;
    cv_bridge::CvImagePtr imagePtr = cv_bridge::toCvCopy(image_msg);


    cv::Mat classified_edges = edgeDetectionClassification(imagePtr->image);

    // find starting points and follow continues lines
    // creating Line objects
    for (int u=1; u < classified_edges.rows-1; u++) {
        for (int v=1; v < classified_edges.cols-1; v++) {
            uint8_t clsif_org = classified_edges.at<uint8_t>(u,v);
            if ( ! clsif_org )
                continue;

            // get only classification without direction:
            uint8_t clsif = 0x0f & clsif_org;

            bool has_neighbour = false;
            switch (clsif) {
                case V:
                    if (
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v))   ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v-1)) ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v+1))
                    )
                        has_neighbour = true;
                    break;
                case D1:
                    if (
                        clsif == (0x0f & classified_edges.at<uint8_t>(u,v+1)) ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v)) ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v+1))
                    )
                        has_neighbour = true;
                    break;
                case H:
                    if (
                        clsif == (0x0f & classified_edges.at<uint8_t>(u,v-1))   ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v-1)) ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v-1))
                    )
                        has_neighbour = true;
                    break;
                case D2:
                    if (
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v)) ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u,v-1)) ||
                        clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v-1))
                    )
                        has_neighbour = true;
                    break;
            }
            if ( has_neighbour )
                continue;

            // no neighbour found means start pixel
            pnt start(v,u);
            // follow line to its end
            do {
                classified_edges.at<uint8_t>(u,v) = 0;
                has_neighbour = false;
                switch (clsif) {
                    case V:
                        if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v)) ) {
                            u++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v+1)) ) {
                            u++;
                            v++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v-1)) ) {
                            u++;
                            v--;
                            has_neighbour = true;
                        }
                        break;
                    case D1:
                        if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v)) ) {
                            u++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u,v-1)) ) {
                            v--;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v-1)) ) {
                            u++;
                            v--;
                            has_neighbour = true;
                        }
                        break;
                    case D2:
                        if ( clsif == (0x0f & classified_edges.at<uint8_t>(u,v+1)) ) {
                            v++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v)) ) {
                            u++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v+1)) ) {
                            u++;
                            v++;
                            has_neighbour = true;
                        }
                        break;
                    case H:
                        if ( clsif == (0x0f & classified_edges.at<uint8_t>(u,v+1)) ) {
                            v++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u+1,v+1)) ) {
                            u++;
                            v++;
                            has_neighbour = true;
                        } else if ( clsif == (0x0f & classified_edges.at<uint8_t>(u-1,v+1)) ) {
                            u--;
                            v++;
                            has_neighbour = true;
                        }
                        break;
                }
            } while ( has_neighbour );

            Line l(start, pnt(v,u), clsif_org);
            if ( l.length > 5.0 ) {
                all_lines.push_back(l);
                if (clsif == D1) {
                    if ( clsif_org & 0x10 ) {
                        right_D1_edges.push_back(l);
                    } else {
                        left_D1_edges.push_back(l);
                    }
                }
                if (clsif == D2) {
                    if ( clsif_org & 0x10 ) {
                        right_D2_edges.push_back(l);
                    } else {
                        left_D2_edges.push_back(l);
                    }
                }
            }

            // reset to original coords, so for loop continues where it left off
            u = start.y;
            v = start.x;
        }
    }


    // match pairs and create LineMarkin objects
    for(const Line& a : left_D1_edges) {
        for(auto it = right_D1_edges.begin(); it != right_D1_edges.end(); it++ ) {
            const Line b = *it;
            if (
                ( a.start.y - 10 < b.start.y && b.start.y < a.start.y + 10 ) &&
                ( a.start.x < b.start.x && b.start.x < a.start.x + 25 )
            ) {
                markings_found.push_back(LineMarking(a,b, D1));
                left_D1_edges.erase(it);
                break;
            }
        }
    }

    for(const Line& a : left_D2_edges) {
        for(auto it = right_D2_edges.begin(); it != right_D2_edges.end(); it++ ) {
            const Line b = *it;
            if (
                ( a.start.y - 10 < b.start.y && b.start.y < a.start.y + 10 ) &&
                ( a.start.x < b.start.x && b.start.x < a.start.x + 25 )
            ) {
                markings_found.push_back(LineMarking(a,b, D2));
                left_D1_edges.erase(it);
                break;
            }
        }
    }

    cv::Mat displayLines = cv::Mat::zeros(cv::Size(imagePtr->image.cols, imagePtr->image.rows), CV_8U);

    for(const Line& l : all_lines) {
        cv::line(displayLines, l.start, l.end, cv::Scalar(40));
    }

    for(const LineMarking& a : markings_found) {

        std::vector<std::vector<cv::Point>> contourVec;
        contourVec.push_back(a.contour);
        cv::drawContours(displayLines, contourVec, 0, cv::Scalar(128));

        cv::line(displayLines, a.middle->start, a.middle->end, cv::Scalar(255));
    }


    cv_bridge::CvImage image_displayLines (std_msgs::Header(), "mono8", displayLines);;
    image_pub->publish(image_displayLines.toImageMsg());

    ros::Time endtime = ros::Time::now();
    ros::Duration time_spend = endtime - starttime;
    ROS_DEBUG("run took: %fs", time_spend.toSec());
}


// node main loop
int main(int argc, char **argv) {
    ros::init(argc, argv, "lane_marker_detection");

    ros::NodeHandle nh;
    ros::NodeHandle private_nh("~");

    /*
     * subscribe to camera image creation event
     */
    ros::Subscriber sub = nh.subscribe("/img/gray", 10, callback_image);

    /*
     * advertise publisher topics
     */
    ros::Publisher image_publisher = private_nh.advertise<sensor_msgs::Image>("/img/temp", 2);
    image_pub = &image_publisher;

    /*
     * start the node
     */
    ros::console::set_logger_level(ROSCONSOLE_DEFAULT_NAME, ros::console::levels::Debug);
    ros::spin();
    return 0;
}
