#include <ros/ros.h>
#include <sensor_msgs/Image.h>
#include <sensor_msgs/image_encodings.h>
#include <yaml-cpp/yaml.h>
#include <opencv2/opencv.hpp>
#include <cv_bridge/cv_bridge.h>


// helper to convert vercor of vectors to a cv::mat
// stolen from https://stackoverflow.com/a/29021835/14157785
template<typename _Tp> static cv::Mat toMat(const std::vector<std::vector<_Tp>> vecIn) {
    cv::Mat_<_Tp> matOut(vecIn.size(), vecIn.at(0).size());
    for (int i = 0; i < matOut.rows; ++i)
        for (int j = 0; j < matOut.cols; ++j)
            matOut(i, j) = vecIn.at(i).at(j);
    return matOut;
}

// global variables needed by callback
ros::Publisher* pub_colorImage;
ros::Publisher* pub_grayImage;
cv::Mat rectifyMapX, rectifyMapY;
cv::Rect ROI;


void callback_undistort_image(sensor_msgs::Image original) {
    ros::Time starttime = ros::Time::now();
    cv::Mat undistortedImage;

    // convert from ROS msg-type to opencv matrix
    cv_bridge::CvImagePtr imagePtr = cv_bridge::toCvCopy(original);

    // apply the calculated rectification maps to undistort the image
    cv::remap(imagePtr->image, undistortedImage, rectifyMapX, rectifyMapY, cv::INTER_LINEAR);

    // crop relevant section from image
    undistortedImage = undistortedImage(ROI);

    // convert to gray
    cv::Mat undistortedImage_gray;
    cv::cvtColor(undistortedImage, undistortedImage_gray, CV_RGB2GRAY);


    // publish images
    cv_bridge::CvImage colorImage(std_msgs::Header(), "rgb8", undistortedImage);
    pub_colorImage->publish(colorImage.toImageMsg());
    cv_bridge::CvImage grayImage(std_msgs::Header(), "mono8", undistortedImage_gray);
    pub_grayImage->publish(grayImage.toImageMsg());

    ROS_DEBUG("run took: %fs", (ros::Time::now()-starttime).toSec());
}


// node main loop
int main(int argc, char **argv) {
    ros::init(argc, argv, "img_undistort");

    ros::NodeHandle nh;
    ros::NodeHandle private_nh("~");

    // ToDo: pass this in as a parameter
    std::string configFilePath = "/home/jetbot/workspace/videodrive_ws/tools/calibration/calibration.yaml";

    // read yaml calibration file
    ROS_INFO("Reading calibration parametes from '%s'", configFilePath.c_str());
    YAML::Node full_config = YAML::LoadFile(configFilePath);
    YAML::Node camera_config = full_config["cameras"]["default"];

    auto distortion_YAML = camera_config["intrinsic"]["distortion"].as<std::vector<double>>();
    cv::Mat distortion ( distortion_YAML );
    ROS_INFO_STREAM("using distortion parameters:\n" << distortion);

    auto cameraMatrix_YAML = camera_config["intrinsic"]["matrix"].as<std::vector<std::vector<double>>>();
    cv::Mat cameraMatrix = toMat( cameraMatrix_YAML );
    ROS_INFO_STREAM("using distortion parameters:\n" << cameraMatrix);

    cv::Size imageSize(
        full_config["images"]["size"]["width"].as<int>(),
        full_config["images"]["size"]["height"].as<int>()
    );
    ROS_INFO("using image Size:\n(%i, %i)", imageSize.width, imageSize.height);

    // get scaled camera matrix
    auto scaledCameraMatrix = cv::getOptimalNewCameraMatrix(cameraMatrix, distortion, imageSize, 1, imageSize, &ROI);
    ROS_INFO("using ROI:\n(%i, %i, %i, %i)", ROI.x, ROI.y, ROI.width, ROI.height);

    // calculate undistortion mapping and storing globaly
    cv::initUndistortRectifyMap(cameraMatrix, distortion, cv::Mat(), scaledCameraMatrix, imageSize, CV_16SC2, rectifyMapX, rectifyMapY);


    /*
     * subscribe to camera image creation event
     */
    ros::Subscriber sub = nh.subscribe("/img/raw", 10, callback_undistort_image);

    /*
     * advertise publisher topics
     */
    ros::Publisher grayImage_publisher = private_nh.advertise<sensor_msgs::Image>("/img/gray", 2);
    pub_grayImage = &grayImage_publisher;
    ros::Publisher colorImage_publisher = private_nh.advertise<sensor_msgs::Image>("/img/color", 2);
    pub_colorImage = &colorImage_publisher;

    /*
     * start the node
     */
    ros::console::set_logger_level(ROSCONSOLE_DEFAULT_NAME, ros::console::levels::Info);
    ros::spin();

    return 0;
}
